"""
Module: '_api.speaker' on micropython-v1.14-893-lego learning system hub
"""
# MCU: {'machine': 'LEGO Technic Large Hub with STM32F413xx', 'sysname': 'LEGO Technic Large Hub', 'platform': 'LEGO Learning System Hub', 'nodename': 'LEGO Learning System Hub', 'ver': 'v1.14-893', 'release': '1.14.0', 'name': 'micropython', 'family': 'micropython', 'port': 'LEGO Learning System Hub', 'version': '1.14.0', 'mpy': 517, 'build': '893'}
# Stubber: 1.7.2
from typing import Any


class Speaker():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    def beep(self, *args, **kwargs) -> Any:
        """"""

    def get_volume(self, *args, **kwargs) -> Any:
        """"""

    def play_sound(self, *args, **kwargs) -> Any:
        """"""

    def set_volume(self, *args, **kwargs) -> Any:
        """"""

    def start_beep(self, *args, **kwargs) -> Any:
        """"""

    def start_sound(self, *args, **kwargs) -> Any:
        """"""

    def stop(self, *args, **kwargs) -> Any:
        """"""


class SpeakerSoundProvider():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    def get_canonical_name(self, *args, **kwargs) -> Any:
        """"""

    def get_sound_files(self, *args, **kwargs) -> Any:
        """"""

system : Any ## <class 'System'> = <System object at 20028790>
def wait_for_async(*args, **kwargs) -> Any:
    """"""

