"""
multithreading support.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/_thread.rst.
========================================

.. module:: _thread
   :synopsis: multithreading support

|see_cpython_module| :mod:`python:_thread`.

This module implements multithreading support.

This module is highly experimental and its API is not yet fully settled
and not yet described in this documentation.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver
