from enum import Enum


class Button(Enum):
    """
    Buttons on a brick or remote.
    """
    LEFT_DOWN = ...
    DOWN = ...
    RIGHT_DOWN = ...
    LEFT = ...
    CENTER = ...
    RIGHT = ...
    LEFT_UP = ...
    UP = ...
    BEACON = ...
    RIGHT_UP = ...
